#include <GL/glfw.h>
#include <cmath>
#include <cstdlib>
#include "entity.hpp"
#include "world.hpp"
#include "main.hpp"

using namespace std;

void drawSprite(float x, float y, float r, float h = .5f);
void drawSprite(float x, float y, float r, float h)
{
	glBegin(GL_QUADS);
	glTexCoord2i(0, 0);
	glVertex3f(x - r, h, y - r);
	glTexCoord2i(1, 0);
	glVertex3f(x - r, h, y + r);
	glTexCoord2i(1, 1);
	glVertex3f(x + r, h, y + r);
	glTexCoord2i(0, 1);
	glVertex3f(x + r, h, y - r);
	glEnd();
}

Entity::Entity()
: x(0), y(0), light(0), size(.25), killme(false)
{
	color[0] = 1.f;
	color[1] = 1.f;
	color[2] = 1.f;
	color[3] = 1.f;
}

void Entity::render()
{
	if (light < 0)
		glBlendFunc(GL_ZERO, GL_ONE_MINUS_SRC_COLOR);
	glColor4fv(color);
	drawSprite(x, y, size + size * 0.18 * sin(world.time * 1.4 + x * 3.123 + y * 2.45123));
	if (light < 0)
		glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);
}

void Entity::update()
{
	if (light < -.0001f || light > .0001f)
	{
		if ((world.frame + (int)x + (int)y * 5) % 14 == 0)
		{
			Particle* p = new Particle(x, y, rand() * .015f / RAND_MAX - .0075f, rand() * .015f / RAND_MAX - .0075f, size * (.4f + rand() * .2f / RAND_MAX), .75f, .75f, .75f, 0.9f);
			if (light < 0.f)
				p->light = -.0000001f;
			world.things.push_back(p);
		}
	}
}

Player::Player()
: nrBreakers(10),
  breakerCountdown(0),
  nrLights(5),
  lightsCountdown(0)
{
	light = 1.1f;
}

void Player::render()
{
	Entity::render();
	glColor3f(.75f, .75f, .75f);
	for (float i = 0; i < M_PI * 2; i += M_PI * 2 / 5)
		drawSprite(x + cos(i + world.time * .7) * .5, y + sin(i + world.time * .7) * .5, .2 + .1 * sin(world.time * (.36 + i * .24) + i * .34));
	glColor3f(.375f, .375f, .375f);
	drawSprite(x, y, .75 + .1 * sin(world.time + x * 1.123 + y * .45123), .35f);
	drawSprite(x, y, .75 + .1 * sin(world.time + x * 1.123 + y * .45123), .65f);
	glColor3f(.25f, .25f, .25f);
	drawSprite(x, y, 1.5 + .2 * sin(world.time + x * 1.123 + y * .45123));
}

void Player::update()
{
	const float speed = .05f;

	float xp = 0, yp = 0;
	if (glfwGetKey(GLFW_KEY_UP)) yp += 1.f;
	if (glfwGetKey(GLFW_KEY_DOWN)) yp -= 1.f;
	if (glfwGetKey(GLFW_KEY_LEFT)) xp += 1.f;
	if (glfwGetKey(GLFW_KEY_RIGHT)) xp -= 1.f;

	if (checkCollide(x + xp * speed, y))
		x += xp * speed;
	if (checkCollide(x, y + yp * speed))
		y += yp * speed;

	if (!breakerCountdown)
	{
		if (glfwGetKey(GLFW_KEY_LCTRL) && nrBreakers)
		{
			playSfx("sfx/shoe.wav");
			world.things.push_front(new Breaker(x, y, speed * xp * 2, speed * yp * 2));
			breakerCountdown = 100;
			nrBreakers--;
		}
	}
	else
		breakerCountdown--;

	if (!lightsCountdown)
	{
		if (glfwGetKey(GLFW_KEY_LALT) && nrLights)
		{
			playSfx("sfx/pop.wav");
			Entity* e = new Entity;
			e->light = .85;
			e->x = x;
			e->y = y;
			world.things.push_front(e);
			lightsCountdown = 100;
			nrLights--;
		}
	}
	else
		lightsCountdown--;

	if (world.frame % 6 == 0)
		world.things.push_back(new Particle(x, y, rand() * .04f / RAND_MAX - .02f, rand() * .04f / RAND_MAX - .02f, .15 + rand() * .2 / RAND_MAX, .9f, .9f, .9f, .7f));

	bool suck = false;
	for (list<Entity*>::iterator i = world.things.begin(); i != world.things.end();)
		if ((*i)->light < -.0001f)
		{
			if (hypotf((*i)->x - x, (*i)->y - y) < 1)
			{
				suck = true;
				(*i)->light += .005f;
				if (world.frame % 5 == 0)
				{
					world.things.push_back(new Particle((*i)->x, (*i)->y, (x - (*i)->x) * .02, (y - (*i)->y) * .02, (*i)->size * .75, .75f, .75f, .75f, .5f));
					world.things.back()->light = -.0000001f;
				}
				if ((*i)->light >= -.0001f)
				{
					delete *i;
					i = world.things.erase(i);
				}
				else
					++i;
			}
			else
				++i;
		}
		else if ((*i)->light < .0001f)
			break;
		else
			++i;
	if (suck && world.frame % 40 == 0)
		playSfx("sfx/suck.wav");
}

bool Player::checkCollide(float cx, float cy)
{
	return world.isFree((int)(cx + .75), (int)cy) && 
	       world.isFree((int)(cx - .75), (int)cy) && 
	       world.isFree((int)cx, (int)(cy + .75)) && 
	       world.isFree((int)cx, (int)(cy - .75)) && 
	       world.isFree((int)(cx + .75), (int)(cy + .75)) && 
	       world.isFree((int)(cx - .75), (int)(cy + .75)) && 
	       world.isFree((int)(cx + .75), (int)(cy - .75)) && 
	       world.isFree((int)(cx - .75), (int)(cy - .75));
}

Breaker::Breaker(float x, float y, float xp, float yp)
: xp(xp), yp(yp)
{
	this->x = x;
	this->y = y;
	color[1] = 0.f;
	color[2] = 0.f;
	light = .2f;
}

void Breaker::update()
{
	x += xp;
	y += yp;
	int ix = (int)x;
	int iy = (int)y;
	if (!world.isFree(ix, iy) || (abs(xp) < .0001f && abs(yp) < .0001f))
	{
		for (int i = iy - 1; i < iy + 2; i++)
			for (int j = ix - 1; j < ix + 2; j++)
			{
				int p = i * world.width + j;
				if (i >= 0 && i < world.height && j >= 0 && j < world.width &&
				    world.tiles[p].isBreakable)
				{
					world.tiles[p].isWall = false;
					world.tiles[p].isBreakable = false;
				}
			}
		// effy
		playSfx("sfx/expl.wav");
		world.things.push_front(new Particle(x - xp, y - yp, 0.f, 0.f, .5f, .0f, .0f, .0f, .15f));
		world.things.front()->light = .7f;
		world.things.push_back(new Particle(x - xp, y - yp, 0.f, 0.f, 1.5f, .7f, .65f, .1f, .4f));
		world.things.push_back(new Particle(x - xp, y - yp, 0.f, 0.f, 2.5f, .2f, .2f, .2f, .8f));
		for (int i = 0; i < 64; i++)
		{
			float xp2 = rand() * .04f / RAND_MAX - .02f;
			float yp2 = rand() * .04f / RAND_MAX - .02f;
			float x2 = x + xp2 * (1.f + rand() * 5.f / RAND_MAX);
			float y2 = y + yp2 * (1.f + rand() * 5.f / RAND_MAX);
			float r, g, b;
			if (rand() & 1)
				r = g = b = rand() * .8f / RAND_MAX;
			else
			{
				r = rand() * .9f / RAND_MAX;
				g = r * .85f;
				b = r * .1f;
			}
			world.things.push_back(new Particle(x2, y2, xp2, yp2, .3 + rand() * .45f / RAND_MAX, r, g, b, .4f + rand() * .5f / RAND_MAX));
		}
		killme = true;
	}
	if (world.frame % 8 == 0)
		world.things.push_back(new Particle(x, y, xp * .5, yp * .5, .5, color[0] * .75, color[1] * .75, color[2] * .75, .7f));
	if ((rand() * 100 / RAND_MAX) % 16 == 0)
		world.things.push_back(new Particle(x, y, xp * .5 + rand() * .02f / RAND_MAX - .01f, yp * .5 + rand() * .02f / RAND_MAX - .01f, size, color[0] * .75, color[1] * .75, color[2] * .75, 1.5f));
}


Particle::Particle(float x, float y, float xp, float yp, float size,
                   float r, float g, float b, float decay)
: xp(xp), yp(yp), decay(decay), start(decay)
{
	this->x = x;
	this->y = y;
	this->size = size;
	color[0] = r;
	color[1] = g;
	color[2] = b;
}

void Particle::update()
{
	decay -= .01f;
	x += xp;
	y += yp;
	if (decay < 0.f || !world.isFree((int)x, (int)y))
		killme = true;
}

void Particle::render()
{
	float orgsize = size;
	float r = color[0], g = color[1], b = color[2];
	color[0] = r * decay / start;
	color[1] = g * decay / start;
	color[2] = b * decay / start;
	size = size * decay * .5 / start + size * .5;
	Entity::render();
	size = orgsize;
	color[0] = r; color[1] = g; color[2] = b;
}
