#include <string>
#include <list>
#include <GL/glfw.h>
#include "main.hpp"
#include "menus.hpp"
#include "world.hpp"

Menu* activeMenu = 0;
MainMenu mainMenu;
PauseMenu pauseMenu;

using namespace std;

Menu::Menu(const string& title)
: title(title), sel(0)
{
}

Menu::~Menu()
{
}

void Menu::key(int k)
{
	switch (k)
	{
		case GLFW_KEY_ESC:
			activeMenu = 0;
			break;
		case GLFW_KEY_UP:
		case GLFW_KEY_LEFT:
			sel--;
			if (sel < 0)
				sel = options.size() - 1;
			break;
		case GLFW_KEY_DOWN:
		case GLFW_KEY_RIGHT:
			sel++;
			if ((unsigned)sel >= options.size())
				sel = 0;
			break;
	}
}

void Menu::draw()
{
	int n = options.size() + 2;
	float h = font.LineHeight();
	glColor3f(1.f, 1.f, 1.f);
	float y = 300 + (h * n) / 2 - h;
	print(font, 400, y, title, true, true);
	y -= 2 * h;
	int c = 0;
	for (list<string>::iterator i = options.begin(); i != options.end(); ++i)
	{
		if (c == sel)
			glColor3f(.5f, 1.f, 1.f);
		else
			glColor3f(1.f, 1.f, 1.f);
		print(font, 400, y, *i, true, true);
		c++;
		y -= h;
	}
}

void Menu::open()
{
	activeMenu = this;
}

MainMenu::MainMenu()
: Menu("Uplighter")
{
	options.push_back("Play");
	options.push_back("Help");
	options.push_back("Exit");
}

void MainMenu::key(int k)
{
	if (k == GLFW_KEY_ESC)
	{
		exitGame = true;
		return;
	}
	Menu::key(k);
	if (k == GLFW_KEY_ENTER)
		switch (sel)
		{
			case 0: // play
				activeMenu = 0;
				startLevel(0);
				break;
			case 1: // help
				helpScreen = true;
				break;
			case 2: // quit
				exitGame = true;
				break;
		}
}

PauseMenu::PauseMenu()
: Menu("Game paused")
{
	options.push_back("Continue");
	options.push_back("Restart");
	options.push_back("Quit");
}

void PauseMenu::key(int k)
{
	if (k == 'P')
		activeMenu = 0;
	Menu::key(k);
	if (k == GLFW_KEY_ENTER)
		switch (sel)
		{
			case 0: // continue
				activeMenu = 0;
				break;
			case 1: // restart
				startLevel(curlev);
				break;
			case 2: // quit
				world.load("title");
				mainMenu.open();
				break;
		}
}

void PauseMenu::open()
{
	sel = 0;
	Menu::open();
}
