#include <cstdio>
#include <cmath>
#include "world.hpp"
#include "main.hpp"

using namespace std;

World world;

World::World()
: width(0), height(0),
  tiles(0), player(0)
{
}

World::~World()
{
	unload();
}

void World::unload()
{
	delete[] tiles;
	delete[] light;
	while (!things.empty())
	{
		delete things.front();
		things.pop_front();
	}
	player = 0;
}

GLuint tex = 0;
GLuint tex2 = 0;
GLuint tex3 = 0;

bool World::load(const string& filename)
{
	unload();

	string tganame = "lvls/" + filename + ".tga";
	GLFWimage map;
	if (!glfwReadImage(tganame.c_str(), &map, 0))
	{
		printf("Could not load map \"%s\"", tganame.c_str());
		return false;
	}

	width = map.Width;
	height = map.Height;
	tiles = new Tile[width * height];

	light = new float[(width + 1) * (height + 1)];
	for (int i = 0; i < (width + 1) * (height + 1); i++)
		light[i] = 0;
	percent = 0;
	goalreached = false;

	unsigned char* m = map.Data;
	for (int y = 0; y < height; y++)
		for (int x = width - 1; x >= 0; x--)
		{
			int p = x + width * y;

			// geometry (stored in R)
			tiles[p].isOut = *m == 0;
			tiles[p].isWall = *m > 100;
			tiles[p].isBreakable = *m > 100 && *m < 250;
			m++;

			// objects
			if (!tiles[p].isOut && !tiles[p].isWall && *m != 0)
			{
				if (*m == 255) // player
				{
					player = new Player;
					player->x = x + .5f;
					player->y = y + .5f;
					things.push_front(player);
				}
				else if (*m > 63) // light
				{
					Entity* e = new Entity;
					e->x = x + .5f;
					e->y = y + .5f;
					if (*m > 127)
						e->light = (*m - 127) / 127.f;
					else
						e->light = (*m - 127) / 63.f;
					things.push_front(e);
				}
			}
			m++;

			// some options
			if (x == width - 1)
				targetpercent = *m;
			if (x == width - 2 && player)
				player->nrBreakers = *m;
			if (x == width - 3 && player)
				player->nrLights = *m;
			m++;
		}

	glfwFreeImage(&map);

	if (!tex)
	{
		glGenTextures(1, &tex);
		glBindTexture(GL_TEXTURE_2D, tex);
		glfwLoadTexture2D("gfx/boards.tga", GLFW_BUILD_MIPMAPS_BIT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	if (!tex2)
	{
		glGenTextures(1, &tex2);
		glBindTexture(GL_TEXTURE_2D, tex2);
		glfwLoadTexture2D("gfx/stone.tga", GLFW_BUILD_MIPMAPS_BIT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	if (!tex3)
	{
		glGenTextures(1, &tex3);
		glBindTexture(GL_TEXTURE_2D, tex3);
		glfwLoadTexture2D("gfx/breakable.tga", GLFW_BUILD_MIPMAPS_BIT);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}

	return true;
}

void World::render()
{
	glDisable(GL_BLEND);
	glDepthMask(1);
	glEnable(GL_TEXTURE_2D);

	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++)
		{
			if (tiles[i * width + j].isOut)
				continue;
			int w = tiles[i * width + j].isWall ? 1 : 0;
			if (tiles[i * width + j].isWall)
			{
				if (tiles[i * width + j].isBreakable)
					glBindTexture(GL_TEXTURE_2D, tex3);
				else
					glBindTexture(GL_TEXTURE_2D, tex2);
			}
			else
				glBindTexture(GL_TEXTURE_2D, tex);		
			float l = tiles[i * width + j].light;
			if (w)
				l = 0;
			glColor3f(l, l, l);
			glBegin(GL_QUADS);
			float tl = light[i * (width + 1) + j];
			glColor3f(tl, tl, tl);
			glTexCoord2i(0, 0);
			glVertex3i(j, w, i);
			float tl2 = light[(i + 1) * (width + 1) + j];
			glColor3f(tl2, tl2, tl2);
			glTexCoord2i(0, 1);
			glVertex3i(j, w, i + 1);
			tl2 = light[(i + 1) * (width + 1) + (j + 1)];
			glColor3f(tl2, tl2, tl2);
			glTexCoord2i(1, 1);
			glVertex3i(j + 1, w, i + 1);
			tl2 = light[i * (width + 1) + (j + 1)];
			glColor3f(tl2, tl2, tl2);
			glTexCoord2i(1, 0);
			glVertex3i(j + 1, w, i);
			glColor3f(tl, tl, tl);
			if (w)
			{
				if (j == 0 || !tiles[i * width + j - 1].isWall)
				{
					tl = light[i * (width + 1) + j] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(0, 0);
					glVertex3i(j, 0, i);
					tl2 = light[(i + 1) * (width + 1) + j] * 1.1;
					glColor3f(tl2, tl2, tl2);
					glTexCoord2i(0, 1);
					glVertex3i(j, 0, i + 1);
					glTexCoord2i(1, 1);
					glVertex3i(j, 1, i + 1);
					glColor3f(tl, tl, tl);
					glTexCoord2i(1, 0);
					glVertex3i(j, 1, i);
				}
				if (j == width - 1 || !tiles[i * width + j + 1].isWall)
				{
					tl = light[i * (width + 1) + j + 1] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(0, 0);
					glVertex3i(j + 1, 0, i);
					glTexCoord2i(0, 1);
					glVertex3i(j + 1, 1, i);
					tl = light[(i + 1) * (width + 1) + j + 1] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(1, 1);
					glVertex3i(j + 1, 1, i + 1);
					glTexCoord2i(1, 0);
					glVertex3i(j + 1, 0, i + 1);
				}
				if (i == 0 || !tiles[(i - 1) * width + j].isWall)
				{
					tl = light[i * (width + 1) + j] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(0, 0);
					glVertex3i(j, 0, i);
					glTexCoord2i(0, 1);
					glVertex3i(j, 1, i);
					tl = light[i * (width + 1) + j + 1] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(1, 1);
					glVertex3i(j + 1, 1, i);
					glTexCoord2i(1, 0);
					glVertex3i(j + 1, 0, i);
				}
				if (i == height - 1 || !tiles[(i + 1) * width + j].isWall)
				{
					tl = light[(i + 1) * (width + 1) + j] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(0, 0);
					glVertex3i(j, 0, i + 1);
					tl2 = light[(i + 1) * (width + 1) + (j + 1)] * 1.1;
					glColor3f(tl, tl, tl);
					glTexCoord2i(0, 1);
					glVertex3i(j + 1, 0, i + 1);
					glTexCoord2i(1, 1);
					glVertex3i(j + 1, 1, i + 1);
					glColor3f(tl, tl, tl);
					glTexCoord2i(1, 0);
					glVertex3i(j, 1, i + 1);
				}
			}
			glEnd();
		}


	static GLuint flare = 0;
	if (!flare)
	{
		glGenTextures(1, &flare);
		glBindTexture(GL_TEXTURE_2D, flare);
		if (!glfwLoadTexture2D("gfx/flare.tga", GLFW_BUILD_MIPMAPS_BIT))
			printf("could not load flare");
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR_MIPMAP_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	}
	glBindTexture(GL_TEXTURE_2D, flare);
	glEnable(GL_BLEND);
	glDepthMask(0);
	glBlendFunc(GL_ONE, GL_ONE_MINUS_SRC_COLOR);

	for (list<Entity*>::iterator i = things.begin(); i != things.end(); ++i)
		(*i)->render();
}

void World::doLights()
{
	// remove some light
	for (int i = 0; i < width * height; i++)
		tiles[i].tmpLight = 0;

	// add light from lights
	int p = 0;
	const float rl = 15;
	for (int i = 0; i < height; i++)
		for (int j = 0; j < width; j++, p++)
			if (!tiles[p].isOut && !tiles[p].isWall)
				for (list<Entity*>::iterator e = things.begin(); e != things.end(); ++e)
					if ((*e)->light < -0.0001f || (*e)->light > 0.0001f)
					{
						if (((*e)->x - j - .5) * ((*e)->x - j - .5) + ((*e)->y - i - .5) * ((*e)->y - i - .5) < rl * rl)
							if (trace((*e)->x, (*e)->y, j + .5, i + .5))
							{
								float dist = 1.0 - hypot((*e)->x - j - .5, (*e)->y - i - .5) / rl;
								tiles[p].tmpLight += (*e)->light * dist * dist * .55;
							}
					}
					else
						break;

	// addblur
	p = width + 1;
	const int rng = 1;
	for (int i = 1; i < height - 1; i++, p += 2)
		for (int j = 1; j < width - 1; j++, p++)
			if (!tiles[p].isOut && !tiles[p].isWall)
			{
				double d = 0;
				double k = 0;
				for (int y = i - rng; y <= i + rng; y++)
					for (int x = j - rng; x <= j + rng; x++)
						if (isFree(x, y)) // && trace(x, y, j, i))
						{
							d += tiles[x + y * width].light;
							k += 1;
						}
				tiles[p].tmpLight += d / k * .7;
			}

	// move light to right place, and "motionblur" it, and calculate percent lit
	percent = 0;
	float c = 0;
	for (int i = 0; i < width * height; i++)
		if (!tiles[i].isWall && !tiles[i].isOut)
		{
			tiles[i].light = tiles[i].tmpLight * .25 + tiles[i].light * .75;
			float l = tiles[i].light * 1.25;
			if (l > 1.f) l = 1.f;
			if (l < 0.f) l = 0.f;
			percent += l;
			c++;
		}
	if (c > 0.0001f)
		percent /= c;
	if (!goalreached && int(100 * percent + .25) >= targetpercent)
	{
		goalreached = true;
		if (player)
			playSfx("sfx/win.wav");
	}
	if (glfwGetKey(GLFW_KEY_F1) && glfwGetKey(GLFW_KEY_F11))
	{
		goalreached = true;
		if (player)
			player->nrBreakers = 99;
	}

	// calculate vertex light
	for (int i = 1; i < height; i++)
		for (int j = 1; j < width; j++)
		{
			int p = i * width + j;
			float l = 0;
			float n = 0;
			if (!tiles[p].isWall && !tiles[p].isOut)
			{
				l += tiles[p].light;
				n += 1.f;
			}
			if (!tiles[p - 1].isWall && !tiles[p - 1].isOut)
			{
				l += tiles[p - 1].light;
				n += 1.f;
			}
			if (!tiles[p - width].isWall && !tiles[p - width].isOut)
			{
				l += tiles[p - width].light;
				n += 1.f;
			}
			if (!tiles[p - 1 - width].isWall && !tiles[p - 1 - width].isOut)
			{
				l += tiles[p - 1 - width].light;
				n += 1.f;
			}
			if (n > .0001f)
				light[p + i] = l / n;
		}
}

bool World::trace(float x1, float y1, float x2, float y2)
{
	int sx = (int)x1;
	int sy = (int)y1;
	int ex = (int)x2;
	int ey = (int)y2;
	float len = hypot(x2 - x1, y2 - y1);
	float leninv = .5f / len;
	float dx = (x2 - x1) * leninv;
	float dy = (y2 - y1) * leninv;
	for (float i = 0; i < len * 2; i++)
	{
		int ix = (int)x1;
		int iy = (int)y1;
		if (!isFree(ix, iy) && !(ix == sx && iy == sy) && !(ix == ex && iy == ey))
			return false;
		x1 += dx;
		y1 += dy;
	}
	return true;
}

bool World::isFree(int x, int y)
{
	if (x < 0 || y < 0 || x >= width || y >= height)
		return false;
	int p = x + y * width;
	return !tiles[p].isOut && !tiles[p].isWall;
}

void World::update()
{
	for (list<Entity*>::iterator i = things.begin(); i != things.end();)
	{
		(*i)->update();
		if ((*i)->killme)
		{
			delete *i;
			i = things.erase(i);
		}
		else
			++i;
	}
}
